/*
** NetXMS - Network Management System
** Copyright (C) 2003-2020 Victor Kirhenshtein
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** File: selfmon.cpp
**
**/

#include "nxcore.h"
#include <gauge_helpers.h>

#define DEBUG_TAG _T("statcoll")

/**
 * Collection interval
 */
#define QUEUE_STATS_COLLECTION_INTERVAL   5

/**
 * Period for calculating moving average
 */
#define QUEUE_STATS_AVERAGE_PERIOD        900

/**
 * Externals
 */
extern ObjectQueue<SyslogMessage> g_syslogProcessingQueue;
extern ObjectQueue<SyslogMessage> g_syslogWriteQueue;
extern ObjectQueue<WindowsEvent> g_windowsEventProcessingQueue;
extern ObjectQueue<WindowsEvent> g_windowsEventWriterQueue;
extern ThreadPool *g_dataCollectorThreadPool;
extern ThreadPool *g_pollerThreadPool;
extern ThreadPool *g_schedulerThreadPool;

int64_t GetEventLogWriterQueueSize();
int64_t GetEventProcessorQueueSize();

/**
 * Internal queue statistic
 */
static StringObjectMap<Gauge64> s_queues(Ownership::True);
static Mutex s_queuesLock;

/**
 * Add queue to collector (using Queue object)
 */
inline void AddQueueToCollector(const TCHAR *name, Queue *queue)
{
   if (queue != nullptr)
   {
      Gauge64 *gauge = new GaugeQueueLength(name, queue, QUEUE_STATS_COLLECTION_INTERVAL, QUEUE_STATS_AVERAGE_PERIOD);
      s_queues.set(name, gauge);
   }
   else
   {
      nxlog_debug_tag(DEBUG_TAG, 2, _T("Queue %s does not exist"));
   }
}

/**
 * Add queue to collector (using thread pool object)
 */
inline void AddQueueToCollector(const TCHAR *name, ThreadPool *pool)
{
   if (pool != nullptr)
   {
      Gauge64 *gauge = new GaugeThreadPoolRequests(name, pool, QUEUE_STATS_COLLECTION_INTERVAL, QUEUE_STATS_AVERAGE_PERIOD);
      s_queues.set(name, gauge);
   }
   else
   {
      nxlog_debug_tag(DEBUG_TAG, 2, _T("Thread pool %s does not exist"));
   }
}

/**
 * Add queue to collector (using callback function)
 */
inline void AddQueueToCollector(const TCHAR *name, int64_t (*function)())
{
   Gauge64 *gauge = new GaugeFunction(name, function, QUEUE_STATS_COLLECTION_INTERVAL, QUEUE_STATS_AVERAGE_PERIOD);
   s_queues.set(name, gauge);
}

/**
 * Get total size of all DB writer queues
 */
static int64_t GetTotalDBWriterQueueSize()
{
   return GetIDataWriterQueueSize() + GetRawDataWriterQueueSize() + g_dbWriterQueue.size();
}

/**
 * Callback for updating queue length gauges
 */
static EnumerationCallbackResult UpdateGauge(const TCHAR *key, const void *value, void *context)
{
   const_cast<Gauge64*>(static_cast<const Gauge64*>(value))->update();
   return _CONTINUE;
}

/**
 * Callback for logging queue lenth stats
 */
static EnumerationCallbackResult LogStats(const TCHAR *key, const void *value, void *context)
{
   nxlog_debug_tag(_T("statcoll.output"), 8, _T("Queue name: %s; Cur: %d; Min: %d; Max: %d; Avg: %f"),
      key,
      const_cast<Gauge64*>(static_cast<const Gauge64*>(value))->getCurrent(),
      const_cast<Gauge64*>(static_cast<const Gauge64*>(value))->getMin(),
      const_cast<Gauge64*>(static_cast<const Gauge64*>(value))->getMax(),
      const_cast<Gauge64*>(static_cast<const Gauge64*>(value))->getAverage());
   return _CONTINUE;
}

/**
 * Statistics collection thread
 */
void ServerStatCollector()
{
   ThreadSetName("StatCollector");

   s_queuesLock.lock();
   AddQueueToCollector(_T("DataCollector"), g_dataCollectorThreadPool);
   AddQueueToCollector(_T("DBWriter.IData"), GetIDataWriterQueueSize);
   AddQueueToCollector(_T("DBWriter.Other"), &g_dbWriterQueue);
   AddQueueToCollector(_T("DBWriter.RawData"), GetRawDataWriterQueueSize);
   AddQueueToCollector(_T("DBWriter.Total"), GetTotalDBWriterQueueSize);
   AddQueueToCollector(_T("EventLogWriter"), GetEventLogWriterQueueSize);
   AddQueueToCollector(_T("EventProcessor"), GetEventProcessorQueueSize);
   AddQueueToCollector(_T("NodeDiscoveryPoller"), GetDiscoveryPollerQueueSize);
   AddQueueToCollector(_T("Poller"), g_pollerThreadPool);
   AddQueueToCollector(_T("Scheduler"), g_schedulerThreadPool);
   AddQueueToCollector(_T("SyslogProcessor"), &g_syslogProcessingQueue);
   AddQueueToCollector(_T("SyslogWriter"), &g_syslogWriteQueue);
   AddQueueToCollector(_T("TemplateUpdater"), &g_templateUpdateQueue);
   AddQueueToCollector(_T("WindowsEventProcessor"), &g_windowsEventProcessingQueue);
   AddQueueToCollector(_T("WindowsEventWriter"), &g_windowsEventWriterQueue);
   s_queuesLock.unlock();

   int counter = 0;
   nxlog_debug_tag(DEBUG_TAG, 1, _T("Server statistic collector thread started"));
   while(!SleepAndCheckForShutdown(QUEUE_STATS_COLLECTION_INTERVAL))
   {
      s_queuesLock.lock();
      s_queues.forEach(UpdateGauge, nullptr);
      counter++;
      if ((counter % (60 / QUEUE_STATS_COLLECTION_INTERVAL)) == 0)
      {
         counter = 0;
         s_queues.forEach(LogStats, nullptr);
      }
      s_queuesLock.unlock();
   }
   nxlog_debug_tag(DEBUG_TAG, 1, _T("Server statistic collector thread stopped"));
}

/**
 * Get queue statistic for internal parameter
 */
DataCollectionError GetQueueStatistic(const TCHAR *parameter, StatisticType type, TCHAR *value)
{
   TCHAR name[MAX_GAUGE_NAME_LEN];
   if (!AgentGetParameterArg(parameter, 1, name, MAX_GAUGE_NAME_LEN))
      return DCE_NOT_SUPPORTED;

   s_queuesLock.lock();

   Gauge64 *stat = s_queues.get(name);
   if (stat == nullptr)
   {
      s_queuesLock.unlock();
      return DCE_NO_SUCH_INSTANCE;
   }

   if (stat->isNull())
   {
      s_queuesLock.unlock();
      return DCE_COLLECTION_ERROR;
   }

   switch(type)
   {
      case StatisticType::AVERAGE:
         ret_double(value, stat->getAverage());
         break;
      case StatisticType::CURRENT:
         ret_int64(value, stat->getCurrent());
         break;
      case StatisticType::MAX:
         ret_int64(value, stat->getMax());
         break;
      case StatisticType::MIN:
         ret_int64(value, stat->getMin());
         break;
   }
   s_queuesLock.unlock();
   return DCE_SUCCESS;
}

/**
 * DCI cache memory usage calculation callback
 */
static void GetCacheMemoryUsage(NetObj *object, uint64_t *size)
{
   if (object->isDataCollectionTarget())
      *size += static_cast<DataCollectionTarget*>(object)->getCacheMemoryUsage();
}

/**
 * Get amount of memory used by DCI cache
 */
uint64_t GetDCICacheMemoryUsage()
{
   uint64_t dciCache = 0;
   g_idxObjectById.forEach(GetCacheMemoryUsage, &dciCache);
   return dciCache;
}

/**
 * NXSL function GetServerQueueNames
 */
int F_GetServerQueueNames(int argc, NXSL_Value **argv, NXSL_Value **result, NXSL_VM *vm)
{
   s_queuesLock.lock();
   StringList *names = s_queues.keys();
   s_queuesLock.unlock();
   *result = vm->createValue(new NXSL_Array(vm, *names));
   delete names;
   return 0;
}
